from pathlib import Path

import pytest
from ape.api import AccountAPI
from ape.exceptions import SignatureError
from ape.types import AddressType
from eth_utils import add_0x_prefix


def test_data_folder(safes, config):
    assert Path.home() not in safes.data_folder.parents
    assert safes.data_folder == config.DATA_FOLDER / "safe"


def test_init(safe, OWNERS, THRESHOLD, safe_contract):
    assert safe.contract == safe_contract
    assert safe.confirmations_required == THRESHOLD
    assert safe.signers == list(o.address for o in OWNERS)
    assert safe.next_nonce == 0


@pytest.mark.parametrize("mode", ("impersonate", "api", "sign"))
def test_swap_owner(safe, accounts, OWNERS, mode):
    impersonate = mode == "impersonate"
    submit = mode != "api"

    old_owner = safe.signers[0]
    new_owner = accounts[len(OWNERS)]  # replace owner 1 with account N + 1
    assert new_owner.address not in safe.signers
    # NOTE: Since the signers are processed in order, we replace the last account

    prev_owner = safe.compute_prev_signer(old_owner)

    def exec_transaction():
        return safe.contract.swapOwner(
            prev_owner,
            old_owner,
            new_owner,
            sender=safe,
            impersonate=impersonate,
            submit=submit,
        )

    if submit:
        receipt = exec_transaction()

    else:
        # Attempting to execute should raise `SignatureError` and push `safe_tx` to mock client
        size = len(list(safe.client.get_transactions(confirmed=False)))
        assert size == 0

        with pytest.raises(SignatureError):
            exec_transaction()

        pending_txns = list(safe.client.get_transactions(confirmed=False))
        assert len(pending_txns) == 1
        assert len(pending_txns[0].confirmations) >= 1
        safe_tx_hash = add_0x_prefix(f"{pending_txns[0].safe_tx_hash}")

        safe_tx_data = pending_txns[0]
        safe_tx = safe.create_safe_tx(**safe_tx_data.model_dump(by_alias=True, mode="json"))

        # Ensure client confirmations works
        client_confs = list(safe.client.get_confirmations(safe_tx_hash))
        assert len(client_confs) >= 1

        # Ensure API confirmations work
        api_confs = safe.get_api_confirmations(safe_tx)
        assert len(api_confs) >= 1

        # `safe_tx` is in mock client, extract it and execute it successfully this time
        receipt = safe.submit_safe_tx(safe_tx)

    assert receipt.events == [
        safe.contract.RemovedOwner(owner=old_owner),
        safe.contract.AddedOwner(owner=new_owner),
        safe.contract.ExecutionSuccess(),
    ]

    assert old_owner not in safe.signers
    assert new_owner.address in safe.signers


@pytest.mark.parametrize("mode", ("impersonate", "api", "sign"))
def test_add_owner(safe, accounts, OWNERS, mode):
    impersonate = mode == "impersonate"
    submit = mode != "api"

    new_owner = accounts[len(OWNERS)]  # replace owner 1 with account N + 1
    assert new_owner.address not in safe.signers

    def exec_transaction():
        return safe.contract.addOwnerWithThreshold(
            new_owner,
            safe.confirmations_required,
            sender=safe,
            impersonate=impersonate,
            submit=submit,
        )

    if submit:
        receipt = exec_transaction()

    else:
        # Attempting to execute should emit a `SignatureError` and push `safe_tx` to mock client
        assert len(list(safe.client.get_transactions(confirmed=False))) == 0
        with pytest.raises(SignatureError):
            exec_transaction()

        assert len(list(safe.client.get_transactions(confirmed=False))) == 1

        # `safe_tx` is in mock client, extract it and execute it successfully this time
        safe_tx_data = next(safe.client.get_transactions(confirmed=False))
        safe_tx = safe.create_safe_tx(**safe_tx_data.model_dump(by_alias=True, mode="json"))
        receipt = safe.submit_safe_tx(safe_tx)

    assert receipt.events == [
        safe.contract.AddedOwner(owner=new_owner),
        safe.contract.ExecutionSuccess(),
    ]

    assert new_owner.address in safe.signers


@pytest.mark.parametrize("mode", ["impersonate", "api", "sign"])
def test_remove_owner(safe, OWNERS, mode):
    impersonate = mode == "impersonate"
    submit = mode != "api"
    if len(OWNERS) == 1:
        pytest.skip("Can't remove the only owner")

    old_owner = safe.signers[0]
    new_threshold = max(len(OWNERS) - 1, safe.confirmations_required - 1)
    threshold_changed = new_threshold != safe.confirmations_required

    prev_owner = safe.compute_prev_signer(old_owner)

    def exec_transaction():
        return safe.contract.removeOwner(
            prev_owner,
            old_owner,
            # Can't set the threshold to zero or more than the number of owners after removal
            new_threshold,
            sender=safe,
            impersonate=impersonate,
            submit=submit,
        )

    if submit:
        receipt = exec_transaction()

    else:
        # Attempting to execute should emit a `SignatureError` and push `safe_tx` to mock client
        assert len(list(safe.client.get_transactions(confirmed=False))) == 0
        with pytest.raises(SignatureError):
            exec_transaction()

        assert len(list(safe.client.get_transactions(confirmed=False))) == 1

        # `safe_tx` is in mock client, extract it and execute it successfully this time
        safe_tx_data = next(safe.client.get_transactions(confirmed=False))
        safe_tx = safe.create_safe_tx(**safe_tx_data.model_dump(by_alias=True, mode="json"))
        receipt = safe.submit_safe_tx(safe_tx)

    expected_events = [
        safe.contract.RemovedOwner(owner=old_owner),
        safe.contract.ExecutionSuccess(),
    ]
    if threshold_changed:
        expected_events.insert(1, safe.contract.ChangedThreshold(threshold=new_threshold))
    assert receipt.events == expected_events

    assert old_owner not in safe.signers


def test_account_type(safe):
    actual = type(safe)
    assert issubclass(actual, AccountAPI)


def test_safe_account_convert(safe):
    """
    We had a bug where converting safe accounts to AddressType
    would fail.
    """
    convert = safe.conversion_manager.convert
    actual = convert(safe, AddressType)
    assert actual == safe.address



import json
import shutil
import tempfile
from pathlib import Path

import ape
import pytest
from ape.contracts import ContractContainer
from ape.utils import ZERO_ADDRESS, create_tempdir
from ethpm_types import ContractType

from ape_safe import MultiSend
from ape_safe.accounts import SafeAccount

contracts_directory = Path(__file__).parent / "contracts"
TESTS_DIR = Path(__file__).parent.absolute()

# TODO: Test more versions.
VERSIONS = ("1.3.0",)


@pytest.fixture(scope="session", autouse=True)
def config(project):
    cfg = ape.config

    # Ensure we have the safe-contracts dependencies.
    for version in VERSIONS:
        _ = project.dependencies.get_dependency("safe-contracts", version)

    # Ensure we don't persist any .ape data.
    with create_tempdir() as path:
        # First, copy in Safe contracts so we don't download each time.
        dest = path / "dest"
        shutil.copytree(cfg.DATA_FOLDER, dest)
        cfg.DATA_FOLDER = dest
        yield cfg


@pytest.fixture
def data_folder(config):
    return config.DATA_FOLDER / "safe"


@pytest.fixture(scope="session")
def deployer(OWNERS):
    return OWNERS[-1]


@pytest.fixture(scope="session")
def receiver(accounts):
    return accounts[9]


@pytest.fixture(scope="session", params=VERSIONS)
def VERSION(request):
    return request.param


@pytest.fixture(scope="session")
def safe_contracts(project, VERSION):
    dependency = project.dependencies.get_dependency("safe-contracts", VERSION)
    return dependency.project


@pytest.fixture(scope="session")
def SafeSingleton(safe_contracts):
    return safe_contracts.GnosisSafe


@pytest.fixture
def singleton(deployer: SafeAccount, SafeSingleton):
    return deployer.deploy(SafeSingleton)


@pytest.fixture(scope="session")
def SafeProxy(safe_contracts, SafeSingleton):
    Proxy = safe_contracts.GnosisSafeProxy
    IProxy = safe_contracts.IProxy
    # NOTE: Proxy only has a constructor, so we add the rest of it's ABI here for simplified use
    Proxy.contract_type.abi += [IProxy.contract_type.abi[0], *SafeSingleton.contract_type.abi]
    return Proxy


@pytest.fixture(scope="session", params=["1/1", "1/2", "2/2", "2/3", "3/3"])
def MULTISIG_TYPE(request):
    # Param is `M/N`, but encoded as a string for repr in pytest
    return request.param.split("/")


@pytest.fixture(scope="session")
def THRESHOLD(MULTISIG_TYPE):
    M, _ = MULTISIG_TYPE
    return int(M)


@pytest.fixture(scope="session")
def OWNERS(accounts, MULTISIG_TYPE):
    _, N = MULTISIG_TYPE
    return accounts[: int(N)]


@pytest.fixture
def safe_contract(singleton, SafeProxy, OWNERS, THRESHOLD):
    deployer = OWNERS[0]
    safe = deployer.deploy(SafeProxy, singleton)
    safe.setup(
        OWNERS,
        THRESHOLD,
        # no modules
        ZERO_ADDRESS,
        b"",
        # no fallback
        ZERO_ADDRESS,
        # no payment
        ZERO_ADDRESS,
        0,
        ZERO_ADDRESS,
        sender=deployer,
    )
    return safe


@pytest.fixture
def safe_data_file(chain, safe_contract):
    with tempfile.NamedTemporaryFile() as fp:
        file = Path(str(fp.name))
        file.write_text(
            json.dumps(
                {
                    "address": safe_contract.address,
                    "deployed_chain_ids": [chain.provider.chain_id],
                }
            )
        )
        yield file


@pytest.fixture
def safe(safe_data_file):
    return SafeAccount(account_file_path=safe_data_file)


@pytest.fixture
def safes():
    return ape.accounts.containers["safe"]


@pytest.fixture
def token(deployer: SafeAccount):
    text = (contracts_directory / "Token.json").read_text()
    contract = ContractType.model_validate_json(text)
    return deployer.deploy(ContractContainer(contract))


@pytest.fixture
def vault(deployer: SafeAccount, token):
    text = (contracts_directory / "VyperVault.json").read_text()
    vault = ContractContainer(ContractType.model_validate_json(text))
    return deployer.deploy(vault, token)


@pytest.fixture
def foundry(networks):
    with networks.ethereum.local.use_provider("foundry") as provider:
        yield provider


@pytest.fixture
def multisend():
    MultiSend.inject()
    return MultiSend()



import click

from ape_safe._cli.pending import pending
from ape_safe._cli.safe_mgmt import _list, add, all_txns, remove


@click.group(short_help="Manage Safe accounts and view Safe API data")
def cli():
    """
    Command-line helper for managing Safes. You can add Safes to your local accounts,
    or view data from any Safe using the Safe API client.
    """


cli.add_command(_list)
cli.add_command(add)
cli.add_command(remove)
cli.add_command(all_txns)
cli.add_command(pending)



import json
from datetime import datetime
from functools import reduce
from typing import Dict, Iterator, Optional, Union, cast

from ape.types import AddressType, HexBytes, MessageSignature
from ape.utils import USER_AGENT, get_package_version
from eip712.common import SafeTxV1, SafeTxV2

from ape_safe.client.base import BaseSafeClient
from ape_safe.client.mock import MockSafeClient
from ape_safe.client.types import (
    ExecutedTxData,
    OperationType,
    SafeApiTxData,
    SafeDetails,
    SafeTx,
    SafeTxConfirmation,
    SafeTxID,
    SignatureType,
    UnexecutedTxData,
)
from ape_safe.exceptions import (
    ClientResponseError,
    ClientUnsupportedChainError,
    MultisigTransactionNotFoundError,
)
from ape_safe.utils import get_safe_tx_hash, order_by_signer

APE_SAFE_VERSION = get_package_version(__name__)
APE_SAFE_USER_AGENT = f"Ape-Safe/{APE_SAFE_VERSION} {USER_AGENT}"
# NOTE: Origin must be a string, but can be json that contains url & name fields
ORIGIN = json.dumps(dict(url="https://apeworx.io", name="Ape Safe", ua=APE_SAFE_USER_AGENT))
assert len(ORIGIN) <= 200  # NOTE: Must be less than 200 chars

TRANSACTION_SERVICE_URL = {
    # NOTE: If URLs need to be updated, a list of available service URLs can be found at
    # https://docs.safe.global/safe-core-api/available-services.
    # NOTE: There should be no trailing slashes at the end of the URL.
    1: "https://safe-transaction-mainnet.safe.global",
    10: "https://safe-transaction-optimism.safe.global",
    56: "https://safe-transaction-bsc.safe.global",
    100: "https://safe-transaction-gnosis-chain.safe.global",
    137: "https://safe-transaction-polygon.safe.global",
    250: "https://safe-txservice.fantom.network",
    288: "https://safe-transaction.mainnet.boba.network",
    8453: "https://safe-transaction-base.safe.global",
    42161: "https://safe-transaction-arbitrum.safe.global",
    43114: "https://safe-transaction-avalanche.safe.global",
    84531: "https://safe-transaction-base-testnet.safe.global",
    11155111: "https://safe-transaction-sepolia.safe.global",
    81457: "https://transaction.blast-safe.io",
}


class SafeClient(BaseSafeClient):
    def __init__(
        self,
        address: AddressType,
        override_url: Optional[str] = None,
        chain_id: Optional[int] = None,
    ) -> None:
        self.address = address

        if override_url:
            tx_service_url = override_url

        elif chain_id:
            if chain_id not in TRANSACTION_SERVICE_URL:
                raise ClientUnsupportedChainError(chain_id)

            tx_service_url = TRANSACTION_SERVICE_URL[chain_id]

        else:
            raise ValueError("Must provide one of chain_id or override_url.")

        super().__init__(tx_service_url)

    @property
    def safe_details(self) -> SafeDetails:
        response = self._get(f"safes/{self.address}")
        return SafeDetails.model_validate(response.json())

    def get_next_nonce(self) -> int:
        return self.safe_details.nonce

    def _all_transactions(self) -> Iterator[SafeApiTxData]:
        """
        Get all transactions from safe, both confirmed and unconfirmed
        """

        url = f"safes/{self.address}/all-transactions"
        while url:
            response = self._get(url)
            data = response.json()

            for txn in data.get("results"):
                # NOTE: Using construct because of pydantic v2 back import validation error.
                if "isExecuted" in txn:
                    if txn["isExecuted"]:
                        yield ExecutedTxData.model_validate(txn)

                    else:
                        yield UnexecutedTxData.model_validate(txn)

                # else it is an incoming transaction

            url = data.get("next")

    def get_confirmations(self, safe_tx_hash: SafeTxID) -> Iterator[SafeTxConfirmation]:
        url = f"multisig-transactions/{str(safe_tx_hash)}/confirmations"
        while url:
            response = self._get(url)
            data = response.json()
            yield from map(SafeTxConfirmation.model_validate, data.get("results"))
            url = data.get("next")

    def post_transaction(
        self, safe_tx: SafeTx, signatures: Dict[AddressType, MessageSignature], **kwargs
    ):
        tx_data = UnexecutedTxData.from_safe_tx(safe_tx, self.safe_details.threshold)
        signature = HexBytes(
            reduce(
                lambda raw_sig, next_sig: raw_sig
                + (next_sig.encode_rsv() if isinstance(next_sig, MessageSignature) else next_sig),
                order_by_signer(signatures),
                b"",
            )
        )
        post_dict: Dict = {"signature": signature.hex(), "origin": ORIGIN}

        for key, value in tx_data.model_dump(by_alias=True, mode="json").items():
            if isinstance(value, HexBytes):
                post_dict[key] = value.hex()
            elif isinstance(value, OperationType):
                post_dict[key] = int(value)
            elif isinstance(value, datetime):
                # not needed
                continue
            else:
                post_dict[key] = value

        post_dict = {**post_dict, **kwargs}

        if "signatures" in post_dict:
            # Signature handled above.
            post_dict.pop("signatures")

        url = f"safes/{tx_data.safe}/multisig-transactions"
        response = self._post(url, json=post_dict)
        return response

    def post_signatures(
        self,
        safe_tx_or_hash: Union[SafeTx, SafeTxID],
        signatures: Dict[AddressType, MessageSignature],
    ):
        if isinstance(safe_tx_or_hash, (SafeTxV1, SafeTxV2)):
            safe_tx = safe_tx_or_hash
            safe_tx_hash = get_safe_tx_hash(safe_tx)
        else:
            safe_tx_hash = safe_tx_or_hash

        safe_tx_hash = cast(SafeTxID, HexBytes(safe_tx_hash).hex())
        url = f"multisig-transactions/{safe_tx_hash}/confirmations"
        signature = HexBytes(b"".join([x.encode_rsv() for x in order_by_signer(signatures)])).hex()
        try:
            self._post(url, json={"signature": signature})
        except ClientResponseError as err:
            if "The requested resource was not found on this server" in err.response.text:
                raise MultisigTransactionNotFoundError(safe_tx_hash, url, err.response) from err

            raise  # The error from BaseClient we are already raising (no changes)

    def estimate_gas_cost(
        self, receiver: AddressType, value: int, data: bytes, operation: int = 0
    ) -> Optional[int]:
        url = f"safes/{self.address}/multisig-transactions/estimations"
        request: Dict = {
            "to": receiver,
            "value": value,
            "data": HexBytes(data).hex(),
            "operation": operation,
        }
        result = self._post(url, json=request).json()
        gas = result.get("safeTxGas")
        return int(HexBytes(gas).hex(), 16)


__all__ = [
    "ExecutedTxData",
    "MockSafeClient",
    "OperationType",
    "SafeApiTxData",
    "SafeClient",
    "SafeDetails",
    "SafeTx",
    "SafeTxConfirmation",
    "SignatureType",
    "UnexecutedTxData",
]



from collections.abc import Iterator
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Optional, Union, cast

from ape.utils import ZERO_ADDRESS, ManagerAccessMixin
from eth_utils import keccak
from hexbytes import HexBytes

from ape_safe.client.base import BaseSafeClient
from ape_safe.client.types import (
    SafeApiTxData,
    SafeDetails,
    SafeTx,
    SafeTxConfirmation,
    SafeTxID,
    SignatureType,
    UnexecutedTxData,
)
from ape_safe.utils import get_safe_tx_hash

if TYPE_CHECKING:
    from ape.contracts import ContractInstance
    from ape.types import AddressType, MessageSignature


class MockSafeClient(BaseSafeClient, ManagerAccessMixin):
    def __init__(self, contract: "ContractInstance"):
        self.contract = contract
        self.transactions: dict[SafeTxID, SafeApiTxData] = {}
        self.transactions_by_nonce: dict[int, list[SafeTxID]] = {}

    @property
    def safe_details(self) -> SafeDetails:
        slot = keccak(text="fallback_manager.handler.address")
        value = self.provider.get_storage(self.contract.address, slot)
        fallback_address = self.network_manager.ecosystem.decode_address(value[-20:])

        return SafeDetails(
            address=self.contract.address,
            nonce=self.get_next_nonce(),
            threshold=self.contract.getThreshold(),
            owners=self.contract.getOwners(),
            masterCopy=self.contract.masterCopy(),
            modules=self.modules,
            # TODO: Add fallback handler getter
            fallbackHandler=fallback_address,
            guard=self.guard,
            version=self.contract.VERSION(),
        )

    @property
    def guard(self) -> "AddressType":
        return (
            self.contract.getGuard() if "getGuard" in self.contract._view_methods_ else ZERO_ADDRESS
        )

    @property
    def modules(self) -> list["AddressType"]:
        return self.contract.getModules() if "getModules" in self.contract._view_methods_ else []

    def get_next_nonce(self) -> int:
        return self.contract._view_methods_["nonce"]()

    def _all_transactions(
        self,
    ) -> Iterator[SafeApiTxData]:
        for nonce in sorted(self.transactions_by_nonce.keys(), reverse=True):
            for tx in map(self.transactions.get, self.transactions_by_nonce[nonce]):
                if tx:
                    yield tx

    def get_confirmations(self, safe_tx_hash: SafeTxID) -> Iterator[SafeTxConfirmation]:
        tx_hash = cast(SafeTxID, HexBytes(safe_tx_hash).hex())
        if safe_tx_data := self.transactions.get(tx_hash):
            yield from safe_tx_data.confirmations

    def post_transaction(
        self, safe_tx: SafeTx, signatures: dict["AddressType", "MessageSignature"], **kwargs
    ):
        safe_tx_data = UnexecutedTxData.from_safe_tx(safe_tx, self.safe_details.threshold)
        safe_tx_data.confirmations.extend(
            SafeTxConfirmation(
                owner=signer,
                submissionDate=datetime.now(timezone.utc),
                signature=sig.encode_rsv(),
                signatureType=SignatureType.EOA,
            )
            for signer, sig in signatures.items()
        )
        tx_id = cast(SafeTxID, HexBytes(safe_tx_data.safe_tx_hash).hex())
        self.transactions[tx_id] = safe_tx_data
        if safe_tx_data.nonce in self.transactions_by_nonce:
            self.transactions_by_nonce[safe_tx_data.nonce].append(tx_id)
        else:
            self.transactions_by_nonce[safe_tx_data.nonce] = [tx_id]

    def post_signatures(
        self,
        safe_tx_or_hash: Union[SafeTx, SafeTxID],
        signatures: dict["AddressType", "MessageSignature"],
    ):
        for signer, signature in signatures.items():
            safe_tx_id = (
                safe_tx_or_hash
                if isinstance(safe_tx_or_hash, (str, bytes, int))
                else get_safe_tx_hash(safe_tx_or_hash)
            )
            tx_id = cast(SafeTxID, HexBytes(safe_tx_id).hex())
            self.transactions[tx_id].confirmations.append(
                SafeTxConfirmation(
                    owner=signer,
                    submissionDate=datetime.now(timezone.utc),
                    signature=signature.encode_rsv(),
                    signatureType=SignatureType.EOA,
                )
            )

    def estimate_gas_cost(
        self, receiver: "AddressType", value: int, data: bytes, operation: int = 0
    ) -> Optional[int]:
        return None  # Estimate gas normally



import click
import rich
from ape.cli import (
    ConnectedProviderCommand,
    network_option,
    non_existing_alias_argument,
    skip_confirmation_option,
)
from ape.exceptions import ChainError, ProviderNotConnectedError
from eth_typing import ChecksumAddress

from ape_safe._cli.click_ext import SafeCliContext, safe_argument, safe_cli_ctx


@click.command(name="list")
@safe_cli_ctx()
@network_option(default=None)
@click.option("--verbose", help="Show verbose info about each safe", is_flag=True)
def _list(cli_ctx: SafeCliContext, network, provider, verbose):
    """
    Show locally-tracked Safes
    """
    if verbose and network is None:
        cli_ctx.abort("Must use '--network' with '--verbose'.")

    network_ctx = None
    if network is not None:
        network_ctx = network.use_provider(provider.name)
        network_ctx.__enter__()

    try:
        number_of_safes = len(cli_ctx.safes)

        if number_of_safes == 0:
            cli_ctx.logger.warning("No Safes found.")
            return

        header = f"Found {number_of_safes} Safe"
        header += "s:" if number_of_safes > 1 else ":"
        click.echo(header)
        total = len(cli_ctx.safes)

        for idx, safe in enumerate(cli_ctx.safes):
            extras = []
            if safe.alias:
                extras.append(f"alias: '{safe.alias}'")

            output: str = ""
            try:
                extras.append(f"version: '{safe.version}'")
            except (ChainError, ProviderNotConnectedError):
                # Not connected to the network where safe is deployed
                extras.append("version: (not connected)")

            else:
                # NOTE: Only handle verbose if we are connected.

                if verbose:
                    local_signers = safe.local_signers or []
                    if local_signers:
                        local_signers_str = ", ".join([x.alias for x in local_signers if x.alias])
                        if local_signers_str:
                            extras.append(f"\n  local signers: '{local_signers_str}'")

                    extras.append(f"next nonce: '{safe.next_nonce}'")
                    extras_joined = ", ".join(extras)
                    extras_display = f"  {extras_joined}" if extras else ""
                    output = f"  {safe.address}{extras_display}"
                    if idx < total - 1:
                        output = f"{output}\n"

            if not output:
                extras_display = f" ({', '.join(extras)})" if extras else ""
                output = f"  {safe.address}{extras_display}"

            rich.print(output)

    finally:
        if network_ctx:
            network_ctx.__exit__(None)


@click.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@click.argument("address", type=ChecksumAddress)
@non_existing_alias_argument()
def add(cli_ctx: SafeCliContext, ecosystem, network, address, alias):
    """
    Add a Safe to locally tracked Safes
    """
    from ape.types import AddressType

    address = cli_ctx.conversion_manager.convert(address, AddressType)
    safe_contract = cli_ctx.chain_manager.contracts.instance_at(address)
    version_display = safe_contract.VERSION()
    required_confirmations = safe_contract.getThreshold()
    signers_display = "\n    - ".join(safe_contract.getOwners())

    cli_ctx.logger.info(
        f"""Safe Found
    network: {ecosystem.name}:{network.name}
    address: {safe_contract.address}
    version: {version_display}
    required confirmations: {required_confirmations}
    signers:
    - {signers_display}
    """
    )

    if click.confirm("Add safe"):
        cli_ctx.safes.save_account(alias, address)
        cli_ctx.logger.success(f"Safe '{address}' ({alias}) added.")


@click.command()
@safe_cli_ctx()
@safe_argument
@skip_confirmation_option()
def remove(cli_ctx: SafeCliContext, safe, skip_confirmation):
    """
    Stop tracking a locally-tracked Safe
    """

    alias = safe.alias
    address = safe.address

    if skip_confirmation or click.confirm(f"Remove safe {address} ({alias})"):
        cli_ctx.safes.delete_account(alias)
        cli_ctx.logger.success(f"Safe '{address}' ({alias}) removed.")


@click.command(cls=ConnectedProviderCommand)
@safe_cli_ctx()
@click.argument("account")
@click.option("--confirmed", is_flag=True, default=None)
def all_txns(cli_ctx: SafeCliContext, account, confirmed):
    """
    View and filter all transactions for a given Safe using Safe API
    """
    from ape.types import AddressType

    from ape_safe.client import ExecutedTxData

    if account in cli_ctx.account_manager.aliases:
        account = cli_ctx.account_manager.load(account)

    address = cli_ctx.conversion_manager.convert(account, AddressType)

    # NOTE: Create a client to support non-local safes.
    client = cli_ctx.safes.create_client(address)

    for txn in client.get_transactions(confirmed=confirmed):
        if isinstance(txn, ExecutedTxData):
            success_str = "success" if txn.is_successful else "revert"
            click.echo(f"Txn {txn.nonce}: {success_str} @ {txn.execution_date}")
        else:
            click.echo(
                f"Txn {txn.nonce}: pending ({len(txn.confirmations)}/{txn.confirmations_required})"
            )



def test_help(runner, cli):
    result = runner.invoke(cli, "--help", catch_exceptions=False)
    assert result.exit_code == 0, result.output


def test_list_no_safes(runner, cli, no_safes):
    result = runner.invoke(cli, "list", catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "No Safes found" in result.output


def test_list_one_safe(runner, cli, one_safe):
    result = runner.invoke(cli, "list", catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "Found 1 Safe" in result.output
    assert one_safe.address in result.output


def test_list_network_not_connected(runner, cli, one_safe):
    result = runner.invoke(
        cli, ("list", "--network", "ethereum:local:test"), catch_exceptions=False
    )
    assert result.exit_code == 0, result.output
    assert "Found 1 Safe" in result.output
    assert one_safe.address in result.output
    assert "not connected" in result.output


def test_list_network_connected(runner, cli, one_safe):
    result = runner.invoke(
        cli, ("list", "--network", "ethereum:local:foundry"), catch_exceptions=False
    )
    assert result.exit_code == 0, result.output
    assert "Found 1 Safe" in result.output
    assert one_safe.address in result.output
    assert "not connected" not in result.output


def test_add_safe(runner, cli, no_safes, safe, chain):
    result = runner.invoke(
        cli,
        ("add", safe.address, safe.alias, "--network", chain.provider.network_choice),
        catch_exceptions=False,
        input="y\n",
    )
    assert result.exit_code == 0, result.output
    assert "SUCCESS" in result.output, result.output


def test_remove_safe(runner, cli, one_safe, safe):
    result = runner.invoke(cli, ("remove", safe.alias), catch_exceptions=False, input="y\n")
    assert result.exit_code == 0, result.output
    assert "SUCCESS" in result.output, result.output


def test_remove_safe_skip_confirmation(runner, cli, one_safe, safe):
    result = runner.invoke(cli, ("remove", safe.alias, "--yes"), catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "SUCCESS" in result.output, result.output



from datetime import datetime, timezone
from enum import Enum
from typing import NewType, Optional, Union, cast

from ape.types import AddressType, HexBytes
from eip712.common import SafeTxV1, SafeTxV2
from eth_typing import HexStr
from eth_utils import add_0x_prefix
from pydantic import BaseModel, Field

from ape_safe.utils import get_safe_tx_hash

SafeTx = Union[SafeTxV1, SafeTxV2]
SafeTxID = NewType("SafeTxID", str)


class SafeDetails(BaseModel):
    address: AddressType
    nonce: int
    threshold: int
    owners: list[AddressType]
    master_copy: AddressType = Field(alias="masterCopy")
    modules: list[AddressType]
    fallback_handler: AddressType = Field(alias="fallbackHandler")
    guard: AddressType
    version: str


class SignatureType(str, Enum):
    APPROVED_HASH = "APPROVED_HASH"
    EOA = "EOA"
    ETH_SIGN = "ETH_SIGN"


class SafeTxConfirmation(BaseModel):
    owner: AddressType
    submission_date: datetime = Field(alias="submissionDate")
    transaction_hash: Optional[HexBytes] = Field(default=None, alias="transactionHash")
    signature: HexBytes
    signature_type: Optional[SignatureType] = Field(default=None, alias="signatureType")


class OperationType(int, Enum):
    CALL = 0
    DELEGATECALL = 1


class UnexecutedTxData(BaseModel):
    safe: AddressType
    to: AddressType
    value: int
    data: Optional[HexBytes] = None
    operation: OperationType
    gas_token: AddressType = Field(alias="gasToken")
    safe_tx_gas: int = Field(alias="safeTxGas")
    base_gas: int = Field(alias="baseGas")
    gas_price: int = Field(alias="gasPrice")
    refund_receiver: AddressType = Field(alias="refundReceiver")
    nonce: int
    submission_date: datetime = Field(alias="submissionDate")
    modified: datetime
    safe_tx_hash: SafeTxID = Field(alias="safeTxHash")
    confirmations_required: int = Field(alias="confirmationsRequired")
    confirmations: list[SafeTxConfirmation] = []
    trusted: bool = True
    signatures: Optional[HexBytes] = None

    @classmethod
    def from_safe_tx(cls, safe_tx: SafeTx, confirmations_required: int) -> "UnexecutedTxData":
        return cls(
            safe=safe_tx._verifyingContract_,
            submissionDate=datetime.now(timezone.utc),
            modified=datetime.now(timezone.utc),
            confirmationsRequired=confirmations_required,
            safeTxHash=get_safe_tx_hash(safe_tx),
            **safe_tx._body_["message"],
        )

    @property
    def base_tx_dict(self) -> dict:
        return {
            "to": self.to,
            "value": self.value,
            "data": self.data,
            "operation": self.operation,
            "safeTxGas": self.safe_tx_gas,
            "baseGas": self.base_gas,
            "gasPrice": self.gas_price,
            "gasToken": self.gas_token,
            "refundReceiver": self.refund_receiver,
            "nonce": self.nonce,
        }

    def __str__(self) -> str:
        # TODO: Decode data
        data_hex = self.data.hex() if self.data else ""
        if len(data_hex) > 40:
            data_hex = f"{data_hex[:18]}....{data_hex[-18:]}"

        # TODO: Handle MultiSend contract differently
        return f"""Tx ID {self.nonce}
   type: {self.operation._name_}
   from: {self.safe}
     to: {self.to}
  value: {self.value / 1e18} ether
   data: {add_0x_prefix(cast(HexStr, data_hex))}
"""


class ExecutedTxData(UnexecutedTxData):
    execution_date: datetime = Field(alias="executionDate")
    block_number: int = Field(alias="blockNumber")
    transaction_hash: HexBytes = Field(alias="transactionHash")
    executor: AddressType
    is_executed: bool = Field(alias="isExecuted")
    is_successful: bool = Field(alias="isSuccessful")
    eth_gas_price: int = Field(alias="ethGasPrice")
    max_fee_per_gas: Optional[int] = Field(alias="maxFeePerGas")
    max_priority_fee_per_gas: Optional[int] = Field(alias="maxPriorityFeePerGas")
    gas_used: int = Field(alias="gasUsed")
    fee: int
    origin: str
    data_decoded: Optional[dict] = Field(alias="dataDecoded")


SafeApiTxData = Union[ExecutedTxData, UnexecutedTxData]



from collections.abc import Mapping
from typing import TYPE_CHECKING, cast

from eip712.messages import calculate_hash
from eth_utils import to_int

if TYPE_CHECKING:
    from ape.types import AddressType, MessageSignature

    from ape_safe.client.types import SafeTxID


def order_by_signer(
    signatures: Mapping["AddressType", "MessageSignature"]
) -> list["MessageSignature"]:
    # NOTE: Must order signatures in ascending order of signer address (converted to int)
    return list(signatures[signer] for signer in sorted(signatures, key=lambda a: to_int(hexstr=a)))


def get_safe_tx_hash(safe_tx) -> "SafeTxID":
    message_hash = calculate_hash(safe_tx.signable_message)
    return cast("SafeTxID", message_hash.hex())



#!/usr/bin/env python
from setuptools import find_packages, setup

extras_require = {
    "test": [  # `test` GitHub Action jobs uses this
        "pytest>=6.0",  # Core testing package
        "pytest-xdist",  # multi-process runner
        "pytest-cov",  # Coverage analyzer plugin
        "hypothesis>=6.2.0,<7.0",  # Strategy-based fuzzer
        "ape-foundry>=0.8",  # Used as the testing provider
        "ape-solidity>=0.8",  # Needed for compiling the Safe contracts
    ],
    "lint": [
        "black>=24.10.0,<25",  # Auto-formatter and linter
        "mypy>=1.13.0,<2",  # Static type analyzer
        "types-requests",  # Needed for mypy type shed
        "types-setuptools",  # Needed for mypy type shed
        "flake8>=7.1.1,<8",  # Style linter
        "flake8-breakpoint>=1.1.0,<2",  # Detect breakpoints left in code
        "flake8-print>=4.0.1,<5",  # Detect print statements left in code
        "flake8-pydantic",  # For detecting issues with Pydantic models
        "flake8-type-checking",  # Detect imports to move in/out of type-checking blocks
        "isort>=5.13.2,<6",  # Import sorting linter
        "mdformat>=0.7.18,<0.8",  # Docs formatter and linter
        "mdformat-pyproject>=0.0.1",  # Allows configuring in pyproject.toml
    ],
    "release": [  # `release` GitHub Action job uses this
        "setuptools",  # Installation tool
        "wheel",  # Packaging tool
        "twine==3.8.0",  # Package upload tool
    ],
    "dev": [
        "commitizen",  # Manage commits and publishing releases
        "pre-commit",  # Ensure that linters are run prior to committing
        "pytest-watch",  # `ptw` test watcher/runner
        "IPython",  # Console for interacting
        "ipdb",  # Debugger (Must use `export PYTHONBREAKPOINT=ipdb.set_trace`)
    ],
}

# NOTE: `pip install -e .[dev]` to install package
extras_require["dev"] = (
    extras_require["test"]
    + extras_require["lint"]
    + extras_require["release"]
    + extras_require["dev"]
)

with open("./README.md") as readme:
    long_description = readme.read()


setup(
    name="ape-safe",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    description="""ape-safe: Gnosis Safe account plugin for Ape""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="ApeWorX Ltd.",
    author_email="admin@apeworx.io",
    url="https://github.com/ApeWorX/ape-safe",
    include_package_data=True,
    install_requires=[
        "eth-ape>=0.8.14,<0.9",
        "requests>=2.31.0,<3",
        "eip712",  # Use same version as eth-ape
        "click",  # Use same version as eth-ape
        "pydantic",  # Use same version as eth-ape
        "eth-utils",  # Use same version as eth-ape
    ],
    entry_points={
        "ape_cli_subcommands": [
            "ape_safe=ape_safe._cli:cli",
        ],
    },
    python_requires=">=3.9,<4",
    extras_require=extras_require,
    py_modules=["ape_safe"],
    license="Apache-2.0",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"ape_safe": ["py.typed", "manifests/*"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



import shutil
from contextlib import contextmanager

import pytest
from ape.utils import create_tempdir
from click.testing import CliRunner

from ape_safe._cli import cli as CLI


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def cli():
    return CLI


@pytest.fixture
def no_safes(data_folder):
    with _remove_safes(data_folder):
        yield


@pytest.fixture
def one_safe(data_folder, safes, safe):
    with _remove_safes(data_folder):
        safes.save_account(safe.alias, safe.address)
        yield safes.load_account(safe.alias)


@contextmanager
def _remove_safes(data_folder):
    with create_tempdir() as temp_dir:
        dest = temp_dir / "dest"
        shutil.move(data_folder, dest)
        yield
        shutil.move(dest, data_folder)



from importlib.resources import files
from io import BytesIO
from typing import TYPE_CHECKING

from ape import convert
from ape.types import AddressType, HexBytes
from ape.utils import ManagerAccessMixin, cached_property
from eth_abi.packed import encode_packed
from ethpm_types import PackageManifest

from ape_safe.exceptions import UnsupportedChainError, ValueRequired

if TYPE_CHECKING:
    from ape.api import ReceiptAPI, TransactionAPI
    from ape.contracts.base import ContractInstance, ContractTransactionHandler

MULTISEND_CALL_ONLY_ADDRESSES = (
    "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D",  # MultiSend Call Only v1.3.0
    "0xA1dabEF33b3B82c7814B6D82A79e50F4AC44102B",  # MultiSend Call Only v1.3.0 (EIP-155)
)
MULTISEND_CALL_ONLY_MANIFEST = PackageManifest.model_validate_json(
    files("ape_safe").joinpath("manifests/multisend.json").read_text()
)
MULTISEND_CALL_ONLY = MULTISEND_CALL_ONLY_MANIFEST.contract_types["MultiSendCallOnly"]  # type: ignore # noqa: E501


class MultiSend(ManagerAccessMixin):
    """
    Create a sequence of calls to execute at once using ``eth_sendTransaction``
    via the MultiSend contract.

    Usage example::

        from ape_safe import multisend
        from ape import accounts

        # load the safe
        safe = accounts.load("my-safe")

        txn = multisend.MultiSend()
        txn.add(contract.myMethod, *call_args)
        txn.add(contract.myMethod, *call_args)
        ...  # Add as many calls as desired to execute
        txn.add(contract.myMethod, *call_args)
        # or, using a builder pattern:
        txn = multisend.MultiSend()
            .add(contract.myMethod, *call_args)
            .add(contract.myMethod, *call_args)
            ...  # Add as many calls as desired to execute
            .add(contract.myMethod, *call_args)

        # Stage the transaction to publish on-chain
        # NOTE: if not enough signers are available, publish to Safe API instead
        receipt = txn(sender=safe,gas=0)
    """

    def __init__(self) -> None:
        """
        Initialize a new MultiSend session object. By default, there are no calls to make.
        """
        self.calls: list[dict] = []

    @classmethod
    def inject(cls):
        """
        Create the multisend module contract on-chain, so we can use it.
        Must use a provider that supports ``debug_setCode``.

        Usage example::

            from ape_safe.multisend import MultiSend

            @pytest.fixture(scope="session")
            def multisend():
                MultiSend.inject()
                return MultiSend()
        """
        active_provider = cls.network_manager.active_provider
        assert active_provider, "Must be connected to an active network to deploy"

        active_provider.set_code(
            MULTISEND_CALL_ONLY_ADDRESSES[0], MULTISEND_CALL_ONLY.get_runtime_bytecode()
        )

    @cached_property
    def contract(self) -> "ContractInstance":
        for address in MULTISEND_CALL_ONLY_ADDRESSES:
            if self.provider.get_code(address) == MULTISEND_CALL_ONLY.get_runtime_bytecode():
                return self.chain_manager.contracts.instance_at(
                    address, contract_type=MULTISEND_CALL_ONLY
                )

        raise UnsupportedChainError()

    @property
    def handler(self) -> "ContractTransactionHandler":
        return self.contract.multiSend

    def add(
        self,
        call,
        *args,
        value=0,
    ) -> "MultiSend":
        """
        Append a call to the MultiSend session object.

        Raises:
            :class:`InvalidOption`: If one of the kwarg modifiers is not able to be used.

        Args:
            call: :class:`ContractMethodHandler` The method to call.
            *args: The arguments to invoke the method with.
            value: int The amount of ether to forward with the call.
        """
        self.calls.append(
            {
                "operation": 0,
                "target": call.contract.address,
                "value": value or 0,
                "callData": call.encode_input(*args),
            }
        )
        return self

    def _validate_calls(self, **txn_kwargs) -> None:
        required_value = sum(call["value"] for call in self.calls)
        if required_value > 0:
            if "value" not in txn_kwargs:
                raise ValueRequired(required_value)

            value = self.conversion_manager.convert(txn_kwargs["value"], int)

            if required_value < value:
                raise ValueRequired(required_value)

        # NOTE: Won't fail if `value` is provided otherwise (won't do anything either)

    @property
    def encoded_calls(self):
        return [
            encode_packed(
                ["uint8", "address", "uint256", "uint256", "bytes"],
                [
                    call["operation"],
                    call["target"],
                    call["value"],
                    len(call["callData"]),
                    call["callData"],
                ],
            )
            for call in self.calls
        ]

    def __call__(self, **txn_kwargs) -> "ReceiptAPI":
        """
        Execute the MultiSend transaction. The transaction will broadcast again every time
        the ``Transaction`` object is called.

        Raises:
            :class:`UnsupportedChain`: If there is not an instance of MultiSend deployed
              on the current chain at the expected address.

        Args:
            **txn_kwargs: the kwargs to pass through to the transaction handler.

        Returns:
            :class:`~ape.api.transactions.ReceiptAPI`
        """
        self._validate_calls(**txn_kwargs)
        if "operation" not in txn_kwargs and not txn_kwargs.get("impersonate", False):
            txn_kwargs["operation"] = 1
        return self.handler(b"".join(self.encoded_calls), **txn_kwargs)

    def as_transaction(self, **txn_kwargs) -> "TransactionAPI":
        """
        Encode the MultiSend transaction as a ``TransactionAPI`` object, but do not execute it.

        Returns:
            :class:`~ape.api.transactions.TransactionAPI`
        """
        self._validate_calls(**txn_kwargs)
        # NOTE: Will fail using `self.handler.as_transaction` because handler
        #       expects to be called only via delegatecall
        if "operation" not in txn_kwargs and not txn_kwargs.get("impersonate", False):
            txn_kwargs["operation"] = 1
        return self.network_manager.ecosystem.create_transaction(
            receiver=self.handler.contract.address,
            data=self.handler.encode_input(b"".join(self.encoded_calls)),
            **txn_kwargs,
        )

    def add_from_calldata(self, calldata: bytes):
        """
        Decode all calls from a multisend calldata and add them to this MultiSend.

        Args:
            calldata: Calldata encoding the MultiSend.multiSend call
        """
        _, args = self.contract.decode_input(calldata)
        buffer = BytesIO(args["transactions"])
        while buffer.tell() < len(args["transactions"]):
            operation = int.from_bytes(buffer.read(1), "big")
            target = convert(buffer.read(20), AddressType)
            value = int.from_bytes(buffer.read(32), "big")
            length = int.from_bytes(buffer.read(32), "big")
            data = HexBytes(buffer.read(length))
            self.calls.append(
                {
                    "operation": operation,
                    "target": target,
                    "value": value,
                    "callData": data,
                }
            )
